package net.sf.sido;

import java.util.Collection;
import java.util.List;

import net.sf.sido.constraint.ConstraintFactory;
import net.sf.sido.editor.DataEditor;

/**
 * A DataFactory defines a set of schemas (and of types) for SiDO instances.
 */
public interface DataFactory extends DataObjectController {

    /**
     * Creates a new schema within this factory.
     * @param name Schema name
     * @return Created schema
     */
	DataSchema createSchema(String name);

    /**
     * Returns the list of types for all schemas in this factory.
     * @return List of types
     */
	List<DataType> getAllTypes();

    /**
     * Gets the factory used to create constraints for properties
     * @return Factory for constraints
     */
	ConstraintFactory getConstraintFactory();

    /**
     * Looks for an editor that can be used for the given type.
     * @param type Value type
     * @param <T> Value type
     * @return Editor for this type
     * @throws IllegalArgumentException If no editor is defined for this type
     */
	<T> DataEditor<T> getEditor(Class<T> type);

    /**
     * Gets a schema using its name
     * @param name Schema name
     * @return The schema or <code>null</code> if not found
     */
	DataSchema getSchema(String name);

    /**
     * Gets the list of all the schemas within this factory
     * @return List of schemas
     */
	Collection<DataSchema> getSchemas();

    /**
     * Gets the list of all sub-types for the given type
     * @param type Parent type to look the sub-types for
     * @return List of sub-types (can be empty but not null)
     */
	List<DataType> getSubTypes(DataType type);

    /**
     * Looks for a type using its qualified name (schema::type).
     * @param qualifiedName Qualified name of the type (schema::type)
     * @return The type or <code>null</code> if not found
     */
	DataType getType(String qualifiedName);

    /**
     * Registers or overrides an editor a value type
     * @param type Value type
     * @param editor Editor for this type
     * @param <T> Value type
     */
	<T> void registerEditor(Class<T> type, DataEditor<T> editor);

    /**
     * Returns the actual type used for a type alias (for example, <code>string</code> will return {@link String}).
     * @param type Alias
     * @param <T> Type
     * @return Actual type or <code>null</code> if this alias is not defined
     */
	<T> Class<T> forAlias(String type);

    /**
     * Gets the actual type for a type definition. This type definition can either be an {@linkplain #forAlias(String) alias}
     * or a fully qualified class name.
     * @param type Type definition
     * @param <T> Type
     * @return Type
     */
	<T> Class<T> getSimpleType(String type);

    /**
     * Creates a new SiDO instance
     * @param type Full type name
     * @return SiDO instance
     * @throws DOTypeNotFoundException If the type cannot be found
     * @see net.sf.sido.DataType#newInstance()
     * @see #getType(String)
     */
	DataObject newInstance(String type);

    /**
     * Creates a new SiDO instance
     * @param type Full type name
     * @param init true if the instance fields must be initialised with default values
     * @return SiDO instance
     * @throws DOTypeNotFoundException If the type cannot be found
     * @see net.sf.sido.DataType#newInstance(boolean)
     * @see #getType(String)
     */
	DataObject newInstance(String type, boolean init);

    /**
     * Creates a new SiDO instance
     * @param type Full type name
     * @param init true if the instance fields must be initialised with default values
     * @param deep true if the initialisation must follow the children
     * @return SiDO instance
     * @throws DOTypeNotFoundException If the type cannot be found
     * @see net.sf.sido.DataType#newInstance(boolean, boolean)
     * @see #getType(String) 
     */
	DataObject newInstance(String type, boolean init, boolean deep);

    /**
     * Looks for a type using its full name and throws an exception
     * if not found
     * @param name Full type name
     * @return Type
     * @throws DOTypeNotFoundException If the type is not found
     */
    DataType getMandatoryType (String name);

}
