package net.sf.sido;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Map;

import net.sf.sido.array.Array;

import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Period;

/**
 * SiDO. A <code>DataObject</code> is the main interface in SiDO. They are never instantiated directly
 * but created using {@link net.sf.sido.DataType#newInstance()} methods.
 */
public interface DataObject extends PropertyAccessor, Serializable, Cloneable {

    /**
     * Creates a deep copy of this SiDO. Each DataObject child is itself {@linkplain #copy() copied}.
     * @return Copy
     */
	DataObject copy();

	/**
	 * Accessor used for Velocity-based access
     * @param name Name of the property to get
     * @return Value for this property
	 */
	Object get(String name);

    /**
     * Gets an {@link Array} property
     * @param property Path to the property
     * @param <T> Type of element in the array
     * @return Array or <code>null</code> if not set
     */
	<T> Array<T> getArray(String property);

    /**
     * Adds an element to an array. If the array is not set, it is created and set before.
     * @param property Path to the property
     * @param elements List of elements to add to the array
     * @param <T> Type of element in the array
     * @return This SiDO
     */
	<T> DataObject add(String property, T... elements);

    /**
     * Gets an array as an array of SiDO.
     * @param property Path to the property
     * @return Array or <code>null</code> if not set
     */
	Array<DataObject> getDOArray(String property);

    /**
     * Gets a boolean property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Boolean getBoolean(String property);

    /**
     * Gets a byte property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Byte getByte(String property);

    /**
     * Gets a SiDO property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	DataObject getDataObject(String property);

    /**
     * Gets a date/time property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	DateTime getDateTime(String property);

    /**
     * Gets a BigDecimal property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	BigDecimal getDecimal(String property);

    /**
     * Gets a double property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Double getDouble(String property);

    /**
     * Gets a duration property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Duration getDuration(String property);

    /**
     * Gets the factory that is associated with this SiDO.
     * @return A factory
     */
	DataFactory getFactory();

    /**
     * Gets a float property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Float getFloat(String property);

    /**
     * Gets a integer property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Integer getInteger(String property);

    /**
     * Gets a long property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Long getLong(String property);

    /**
     * Gets a property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     * @param <T> Type of the property value
     */
	<T> T getObject(String property);

    /**
     * Gets any link to this SiDO parent. If this SiDO has been set as the property
     * value of a parent item and if this property has been declared as "reverse", then
     * the parent link will be filled.
     * @return The link or <code>null</code> if no link is defined
     */
	ParentLink getParentLink();

    /**
     * Gets a period property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Period getPeriod(String property);

    /**
     * Gets a short property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	Short getShort(String property);

	/**
	 * Returns a unique ID for this instance inside its type. This ID has been created
     * by the associated factory and is unique for this factory and the SiDO's type.
     * @return ID
	 */
	int getSidoId();

    /**
     * Gets a string property
     * @param property Path to the property
     * @return Value or <code>null</code> if not set
     */
	String getString(String property);

    /**
     * Gets the type of this SiDO.
     * @return Type
     */
	DataType getSidoType();

    /**
     * Gets a property using the property definition
     * @param property Property definition
     * @return Value or <code>null</code> if not set
     * @param <T> Type of the property
     * @see net.sf.sido.DataType#getProperties()
     * @see net.sf.sido.DataType#getProperty(String)
     */
	<T> T getValue(Property<T> property);

    /**
     * Returns an immutable copy of this SiDO.
     * @return SiDO
     */
	DataObject immutable();

    /**
     * Initialises all values of this SiDO using default values. No deep initialisation is performed.
     * @see #init(boolean)
     */
	void init();

    /**
     * Initialises all values of this SiDO using default values. If <code>deep</code> is <code>true</code>, then
     * all SiDO children are also initialised recursively if they are not <i>nullable</i>.
     * @param deep <code>true</code> to initialise all non nullable children recursively
     */
	void init(boolean deep);

    /**
     * Checks if this SiDO is immutable. If immutable, so value can be changed on it.
     * @return <code>true</code> if this SiDO is immutable, <code>false</code> otherwise.
     */
	boolean isImmutable();

    /**
     * Sets an array as a property
     * @param property Path to this property
     * @param value New value to set
     * @param <T> Type of element for the array
     * @return This SiDO
     */
	<T> DataObject setArray(String property, Array<T> value);

    /**
     * Sets an array as a property using a collection
     * @param property Path to this property
     * @param value New value to set
     * @param <T> Type of element for the array
     * @return This SiDO
     */
	<T> DataObject setArray(String property, Collection<T> value);

    /**
     * Sets a property using value(s) given as strings. Multiple values are used only for array properties: using
     * more than 1 value for a single property is illegal and using none means settings the value to <code>null</code>.
     * Value conversions are performed using editors defined at factory level.
     * @param property Path to the property
     * @param values List of values
     * @return This SiDO
     * @see DataFactory#getEditor(Class)
     */
	DataObject setAsString(String property, String... values);

    /**
     * Sets a boolean as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setBoolean(String property, Boolean value);

    /**
     * Sets a byte as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setByte(String property, Byte value);

    /**
     * Sets a SiDO as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setDataObject(String property, DataObject value);

    /**
     * Sets a date/time as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setDateTime(String property, DateTime value);

    /**
     * Sets a decimal as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setDecimal(String property, BigDecimal value);

    /**
     * Sets a double as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setDouble(String property, Double value);

    /**
     * Sets a duration as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setDuration(String property, Duration value);

    /**
     * Sets a float as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setFloat(String property, Float value);

    /**
     * Sets an integer as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setInteger(String property, Integer value);

    /**
     * Sets a long as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setLong(String property, Long value);

    /**
     * Sets an object as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     * @param <T> Type of this value
     */
	<T> DataObject setObject(String property, T value);

    /**
     * Sets the parent for this SiDO. This method should be used only by the SiDO internal API!
     * @param link Link to set
     */
	void setParentLink(ParentLink link);

    /**
     * Sets a period as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setPeriod(String property, Period value);

    /**
     * Sets a short as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setShort(String property, Short value);

    /**
     * Sets a string as property
     * @param property Path to the property
     * @param value New value to set
     * @return This SiDO
     */
	DataObject setString(String property, String value);

    /**
     * Sets a property using its definition
     * @param property Property definition
     * @param value New value to set
     * @return This SiDO
     * @param <T> Type for the property value
     */
	<T> DataObject setValue(Property<T> property, T value);

    /**
     * Validates the SiDO using property definitions and constraints.
     * @return Result of the validation
     */
	ValidationResults validate();

    /**
     * Visits all property values of this SiDO and applies a visitor to each value recursively.
     * @param visitor Visitor to apply
     */
	void visit(DataObjectVisitor visitor);

    /**
     * Visits all property values of this SiDO and applies a visitor to each value recursively.
     * @param visitor Visitor to apply
     * @param context Visitor context to carry along
     */
	void visit(DataObjectVisitor visitor, VisitorContext context);

	/**
	 * Copies a set of properties from a {@link DataObject} to this one.
	 * @param from DataObject to copy properties from
	 * @param copy If <code>true</code>, the properties which are {@link DataObject} or arrays of {@link DataObject} are
	 * themselves {@linkplain #copy() copied}. If <code>false</code>, those properties are copied by reference.
	 * @param properties List of properties to be copied.
	 * @return This instance
	 */
	DataObject copy(DataObject from, boolean copy, String... properties);

	/**
	 * Copies a set of properties from a {@link DataObject} to this one. The {@link DataObject} properties are copied by
	 * reference (see {@link #copy(DataObject, boolean, String...)} for have a different behaviour).
	 * @param from DataObject to copy properties from
	 * @param properties List of properties to be copied.
	 * @return This instance
	 * @see #copy(DataObject, boolean, String...)
	 */
	DataObject copy(DataObject from, String... properties);

    /**
     * Sets an enumeration value. Internally, enumerated properties are defined and stored as strings.
     * @param property Path to the property
     * @param value New value to set
     * @param <E> Enumeration type
     * @return This SiDO
     */
	<E extends Enum<E>> DataObject setEnum(String property, E value);

    /**
     * Gets an enumerated property. Internally, enumerated properties are defined and stored as strings.
     * @param property Path to the property.
     * @param enumClass Class of the enumeration.
     * @param <E> Enumeration type
     * @return Value or <code>null</code> if not set
     */
	<E extends Enum<E>> E getEnum(String property, Class<E> enumClass);

    /**
     * Computes the difference between this SiDO and <code>that</code> SiDO.
     *
     * The returned map contains all differences between the two objects, indexed by the property paths.
     * @param that SiDO to compare with
     * @return Differences
     */
    Map<String,Difference> diff (DataObject that);

}
